$link-color: #386da5;
$white: #fff;
$body-background-color: $white;
$sidebar-color: $white;

.bg-contribution {
  background-color: rgba(255, 233, 112, 0.1);
}

.navigation-list-item .navigation-list-child-list {
  display: block !important;
}
