#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['topojson', 'topojson.core']

package_data = \
{'': ['*'],
 'topojson': ['.ipynb_checkpoints/*'],
 'topojson.core': ['.ipynb_checkpoints/*']}

install_requires = \
['numpy', 'shapely']

extras_require = \
{'dev': ['shapely >=1.7a2',
         'geojson',
         'simplification',
         'pyshp',
         'fiona >=1.8.6',
         'geopandas',
         'altair',
         'ipywidgets']}

setup(name='topojson',
      version='1.0rc9',
      description='topojson - a powerful library to encode geographic data as topology in Python!🌍',
      author='Mattijn van Hoek',
      author_email='mattijn@gmail.com',
      url='https://github.com/mattijn/topojson',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )
